# -*- coding: utf-8 -*-

from bravaweb.response.lib.object import ResponseObject
from bravaweb.response.lib.view import View

from bravaweb.response.css import Css
from bravaweb.response.csv import Csv
from bravaweb.response.html import Html
from bravaweb.response.javascript import JavaScript
from bravaweb.response.jpg import Jpg
from bravaweb.response.json import Json
from bravaweb.response.mp4 import Mp4
from bravaweb.response.pdf import Pdf
from bravaweb.response.png import Png
from bravaweb.response.textplain import TextPlain
from bravaweb.response.webm import Webm
from bravaweb.response.xml import Xml



# -*- coding: utf-8 -*-

# Third Part Imports

# Basics Imports
import configuration

# Utils
from bravaweb.utils import GetHeader

# Security
import bravaweb.security as Security

# Enviroment
from bravaweb.enviroment import Enviroment

# Error Tratament
import traceback

# App Boot Time
from datetime import datetime

boot_time =datetime.now()

async def App(scope, receive, send):

    headers = GetHeader(scope)

    if scope["method"] == "OPTIONS":
        await Security.origin.Options(headers, send)

    elif Security.origin.Static(scope):
        await Security.origin.NotFound(send)

    elif Security.origin.Permitted(headers, scope):

        try:

            envirom = Enviroment(headers, scope, receive, send)
            await envirom.ReceiveBody()
            await envirom.Response()

        except Exception as e:
            print("\n\nError: {}\n\n-- TRACEBACK --\n\n{}\n\n".format(e, traceback.format_exc()))
    else:
        await Security.origin.Forbidden(send)




# DECORATE ALL CALLABLE METHODS IN CLASS


def require(decorator):
    def decorate(cls):
        for attr in cls.__dict__:
            if hasattr(getattr(cls, attr), '__call__'):
                setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls
    return decorate



# -*- coding: utf-8 -*-

from bravaweb.security.methods import *
from bravaweb.security.decorator import require
from bravaweb.utils import ResponseType, ResponseCode, RequestMethod
from bravaweb.response import *

from datetime import datetime
import decimal

import configuration

class Controller(object):

    def __init__(self, envirom):
        self.enviroment = envirom
        super(Controller, self).__init__()

    # TODO VERIFICAR SE ESTA USANDO ELE, O ENTER OU EXIT
    def __del__(self):
        pass

    async def NotAllowed(self):

        content = "405: Method not allowed".encode(configuration.api.encoding)

        _headers = [
            (b'content-type', b'text/html; charset=utf-8'),
            (b"Content-Length", str(len(content)).encode(configuration.api.encoding)),
            (b"Accept-Ranges", b"bytes"),
            (b"X-Frame-Options:", b"Deny")
        ]

        if self.enviroment.origin:
            _headers.append((b"Access-Control-Allow-Origin", self.enviroment.origin.encode(configuration.api.encoding)))

        await self.enviroment.send({
            'type': 'http.response.start',
            'status': ResponseCode.MethodNotAllowed.code,
            'headers': _headers
        })

        await self.enviroment.send({
            'type': 'http.response.body',
            'body': content
        })

    async def Unauthorized(self):

        content = "401: Unauthorized".encode(configuration.api.encoding)

        _headers = [
            (b'content-type', b'text/html; charset=utf-8'),
            (b"Content-Length", str(len(content)).encode(configuration.api.encoding)),
            (b"Accept-Ranges", b"bytes"),
            (b"X-Frame-Options:", b"Deny")
        ]

        if self.enviroment.origin:
            _headers.append((b"Access-Control-Allow-Origin", self.enviroment.origin.encode(configuration.api.encoding)))

        await self.enviroment.send({
            'type': 'http.response.start',
            'status': ResponseCode.Unauthorized.code,
            'headers': _headers
        })

        await self.enviroment.send({
            'type': 'http.response.body',
            'body': content
        })

    async def NotFound(self):

        content = "404: Not Found".encode(configuration.api.encoding)

        _headers = [
            (b'content-type', b'text/html; charset=utf-8'),
            (b"Content-Length", str(len(content)).encode(configuration.api.encoding)),
            (b"Accept-Ranges", b"bytes"),
            (b"X-Frame-Options:", b"Deny")
        ]

        if self.enviroment.origin:
            _headers.append((b"Access-Control-Allow-Origin", self.enviroment.origin.encode(configuration.api.encoding)))

        await self.enviroment.send({
            'type': 'http.response.start',
            'status': ResponseCode.NotFound.code,
            'headers': _headers
        })

        await self.enviroment.send({
            'type': 'http.response.body',
            'body': content
        })

    async def PreconditionFailed(self, errors):

        content = "412: Precondition Failed".encode(configuration.api.encoding)

        _headers = [
            (b'content-type', b'text/html; charset=utf-8'),
            (b"Content-Length", str(len(content)).encode(configuration.api.encoding)),
            (b"Accept-Ranges", b"bytes"),
            (b"X-Frame-Options:", b"Deny")
        ]

        if self.enviroment.origin:
            _headers.append((b"Access-Control-Allow-Origin", self.enviroment.origin.encode(configuration.api.encoding)))

        await self.enviroment.send({
            'type': 'http.response.start',
            'status': ResponseCode.PreconditionFailed.code,
            'headers': _headers
        })

        await self.enviroment.send({
            'type': 'http.response.body',
            'body': content
        })

    async def InternalError(self):

        content = "500: Internal Error".encode(configuration.api.encoding)

        _headers = [
            (b'content-type', b'text/html; charset=utf-8'),
            (b"Content-Length", str(len(content)).encode(configuration.api.encoding)),
            (b"Accept-Ranges", b"bytes"),
            (b"X-Frame-Options:", b"Deny")
        ]

        if self.enviroment.origin:
            _headers.append((b"Access-Control-Allow-Origin", self.enviroment.origin.encode(configuration.api.encoding)))

        await self.enviroment.send({
            'type': 'http.response.start',
            'status': ResponseCode.InternalError.code,
            'headers': _headers
        })

        await self.enviroment.send({
            'type': 'http.response.body',
            'body': content
        })



# -*- coding: utf-8 -*-

import configuration
import re

from bravaweb.utils import ResponseCode


async def Options(headers, send):
    try:
        if headers["origin"] in configuration.api.domains:
            await send({
                'type': 'http.response.start',
                'status': ResponseCode.OK.code,
                'headers': [
                    (b"Access-Control-Allow-Headers",
                     b"Authorization, Content-Type"),
                    (b"Access-Control-Allow-Origin",
                     headers["origin"].encode(configuration.api.encoding)),
                    (b"Access-Control-Allow-Methods",
                     b"GET, POST, OPTIONS, PUT, DELETE"),
                    (b"Accept-Ranges", b"bytes"),
                    (b"X-Frame-Options:", b"Deny")
                ]
            })
        else:
            await send({
                'type': 'http.response.start',
                'status': ResponseCode.Forbidden.code,
                'headers': [
                    (b"X-Frame-Options:", b"Deny")
                ]
            })
        await send({
            'type': 'http.response.body',
            'body': b'',
        })
    except Exception as e:
        print("\n\nOptions Erro:\nbravaweb.security.origin.Options()\n\n")
        raise e


def TestPermissionRule(_var, _key, _rule):
    if _rule == "*":
        return True
    elif not _key in _var:
        return False
    else:
        return bool(re.match(_rule, _var[_key]))


def Permitted(headers, scope):
    permitted = False
    if "origin" in headers and headers["origin"] in configuration.api.domains:
        permitted = True
    if not permitted:
        permitted = any([TestPermissionRule(scope, "path", exc["path"]) and TestPermissionRule(
            headers, "referer", exc["referer"]) for exc in configuration.api.access_exceptions])
    return permitted


def Static(scope):
    return scope["path"] == "/favicon.ico"


async def Forbidden(send):

    await send({
        'type': 'http.response.start',
        'status': ResponseCode.Forbidden.code,
        'headers': [
            (b"X-Frame-Options:", b"Deny")
        ]
    })

    await send({
        'type': 'http.response.body',
        'body': b'403: Forbidden',
    })


async def NotFound(send):

    await send({
        'type': 'http.response.start',
        'status': ResponseCode.NotFound.code,
        'headers': [
            (b"X-Frame-Options:", b"Deny")
        ]
    })

    await send({
        'type': 'http.response.body',
        'body': b'404: Not Found',
    })




import traceback
import configuration
from datetime import datetime

def ValidRequestParams(enviroment, action, klass):

    params = {}
    errors = []
    annotations = action.__annotations__

    # WHILE DECORATOR, GET ENCAPSULATE FUNCTION
    if "return" in annotations:
        while "decorator" in annotations["return"].__name__:
            annotations = annotations["return"].__annotations__

        # GET ANOTATION FUNCTION
        annotations = dict(annotations["return"].__annotations__)

        # SET ENVIROMENT RESPONSE TYPE
        enviroment.response_type = annotations["return"]

        # REMOVE RETURN FROM REQUIRED PARAMS
        try:
            del annotations["return"]
        except Exception as e:
            pass

        # GET REQUIRED PARAM IN REQUEST
        for param, instance in annotations.items():
            if param not in enviroment.request:
                errors.append(
                    "Field [{}] not present in the request".format(param))
            else:
                try:
                    if isinstance(enviroment.request[param], dict):
                        params[param] = instance(**enviroment.request[param])
                    elif instance.__name__ == "bool" and isinstance(enviroment.request[param], str):
                        params[param] = enviroment.request[param].lower() in ['true', '1']
                    elif instance.__name__ == "datetime" and isinstance(enviroment.request[param], str):
                        params[param] = datetime.strptime(enviroment.request[param], configuration.api.date_format)
                    else:
                        params[param] = instance(enviroment.request[param])
                except ValueError:
                    errors.append("Invalid value for [{}] field".format(param))
                except Exception as e:
                    errors.append("Unexpected [{}] field".format(param))

    return params, errors



# -*- coding: utf-8 -*-

# Third Part
import re
import jwt
import json
import urllib.parse

# configuration
import configuration

# Utils
from bravaweb.utils import GetBody
from bravaweb.utils import RequestMethod

# Secutiry
from bravaweb.security.request_params import ValidRequestParams


class Enviroment():

    def __init__(self, headers, scope, receive, send):

        # Init Params
        self.headers = headers
        self.scope = scope
        self.send = send
        self.receive = receive

        # Default Params
        self.origin = headers["origin"] if "origin" in headers else None
        self.remote_ip = scope['client'][0]
        self.browser = headers["user-agent"] if "user-agent" in headers else None
        self.accept_encoding = headers["accept-encoding"] if "accept-encoding" in headers else None
        self.remote = {"ip": self.remote_ip,
                       "browser": self.browser, "accept": self.accept_encoding}
        self.method = RequestMethod(scope["method"])
        self.response_type = None

        # Authorization
        self.authorization = headers["authorization"] if "authorization" in headers else None
        self.auth_token = None
        self.bearer = jwt.decode(headers["authorization"].replace(
            'Bearer ', ''), configuration.api.token, algorithms='HS256') if self.authorization is not None else None

        # Content Length
        self.content_length = int(
            headers['content-length']) if "content-length" in headers else 0

        # Query String
        self.get = {}
        for item in scope["query_string"].decode(configuration.api.encoding).split("&"):
            if "=" in item:
                item_name = urllib.parse.unquote(item.split("=")[0])
                item_value = urllib.parse.unquote(item.split("=")[1])
                if item_name.find("[") >= 0  and item_name.find("]") >= 0:
                    object_name, field_name = item_name.split("[")[0], item_name.split("[")[1].replace("]","")
                    if object_name in self.get:
                        self.get[object_name][field_name] = item_value
                    else:
                        self.get[object_name] = {field_name : item_value}
                else:
                    self.get[item_name] = item_value
            else:
                item_name = urllib.parse.unquote(item)

        # Post
        self.post = {}

        # Validate Route
        self.controller, self.area, self.module, self.action, self.id = None, None, "default", "index", None

        # Route Rule
        RouteRule = [route for route in configuration.api.routes
                     if bool(re.match(route[1], scope["path"]))][0][0]

        # Request Route
        RequestRoute = scope["path"][1:].split("/")
        for index, attr in enumerate(RouteRule.split("/")):
            # If Attribute was defined
            if re.match("\{.*?\}", attr):
                try:
                    self.__setattr__(
                        attr[attr.find("{") + 1:attr.find("}")], RequestRoute[index])
                except IndexError:
                    pass
            # Else Pattern
            elif index == 0:
                self.controller = attr
            elif index == 1:
                self.area = attr
            elif index == 2:
                self.module = attr
            elif index == 3:
                self.action = attr
            elif index == 4:
                self.id = attr

        self.route = ["controllers"]

        if self.controller:
            self.route.append(self.controller)
        if self.area:
            self.route.append(self.area)
        if self.module:
            self.route.append(self.module)

        self.klass = "{}Controller".format(self.route[-1].capitalize())

    async def ReceiveBody(self):
        self.body = await GetBody(self.receive)

        # Post
        if len(self.body) > 0:
            try:
                self.post = json.loads(
                    self.body.decode(configuration.api.encoding))
            except ValueError:
                for item in self.body.decode(configuration.api.encoding).split("&"):
                    if "=" in item:
                        item_name = urllib.parse.unquote(item.split("=")[0])
                        item_value = urllib.parse.unquote(item.split("=")[1])
                        if item_name.find("[") >= 0  and item_name.find("]") >= 0:
                            object_name, field_name = item_name.split("[")[0], item_name.split("[")[1].replace("]","")
                            if object_name in self.post:
                                self.post[object_name][field_name] = item_value
                            else:
                                self.post[object_name] = {field_name : item_value}
                        else:
                            self.post[item_name] = item_value
                    else:
                        item_name = urllib.parse.unquote(item)

        # Union Get and Post Params
        self.request = {**self.get, **self.post}

    async def Response(self):

        # Load Module
        try:
            _module = __import__(".".join(self.route), fromlist=[self.klass])
        except Exception as e:
            raise e

        # Load Class
        try:
            _class = _module.__getattribute__(self.klass)(envirom=self)
        except Exception as e:
            raise e

        # Load Action
        try:
            _action = _class.__getattribute__(self.action)
        except AttributeError:
            # 404 Not Found
            _action = _class.__getattribute__("NotFound")
        except Exception as e:
            raise e

        # Validate Request Action
        try:
            params, errors = ValidRequestParams(self, _action, _class)
            if len(errors) > 0:
                _action = _class.__getattribute__("PreconditionFailed")
                params = {'errors': errors}
        except Exception as e:
            raise e

        # Execute Action
        try:
            await _action(**params)
        except Exception as e:
            raise e



# -*- coding: utf-8 -*-

from enum import Enum


class ResponseCode(Enum):

    OK = 200
    NotFound = 404
    Invalid = 400
    Forbidden = 403
    PreconditionFailed = 412
    InternalError = 500
    Redirect = 302
    Permanent = 301
    Temporary = 307
    MethodNotAllowed = 405
    Unauthorized = 401

    def __str__(self):
        return "%s %s" % (self.value, self.description)

    @property
    def description(self):
        codes = {
            # Informational.
            100: "Continue",
            101: "Switching Protocols",
            102: "Processing",
            103: "Checkpoint",
            122: "Uri Too Long",
            200: "Ok",
            201: "Created",
            202: "Accepted",
            203: "Non Authoritative Info",
            204: "No Content",
            205: "Reset Content",
            206: "Partial Content",
            207: "Multi Status",
            208: "Already Reported",
            226: "Im Used",

            # Redirection.
            300: "Multiple Choices",
            301: "Moved Permanently",
            302: "Found",
            303: "See Other",
            304: "Not Modified",
            305: "Use Proxy",
            306: "Switch Proxy",
            307: "Temporary Redirect",
            308: "Permanent Redirect",

            # Client Error.
            400: "Bad Request",
            401: "Unauthorized",
            402: "Payment Required",
            403: "Forbidden",
            404: "Not Found",
            405: "Method Not Allowed",
            406: "Not Acceptable",
            407: "Proxy Authentication Required",
            408: "Request Timeout",
            409: "Conflict",
            410: "Gone",
            411: "Length Required",
            412: "Precondition Failed",
            413: "Request Entity Too Large",
            414: "Request Uri Too Large",
            415: "Unsupported Media Type",
            416: "Requested Range Not Satisfiable",
            417: "Expectation Failed",
            418: "Im A Teapot",
            421: "Misdirected Request",
            422: "Unprocessable Entity",
            423: "Locked",
            424: "Failed Dependency",
            425: "Unordered Collection",
            426: "Upgrade Required",
            428: "Precondition Required",
            429: "Too Many Requests",
            431: "Header Fields Too Large",
            444: "No Response",
            449: "Retry With",
            450: "Blocked By Windows Parental Controls",
            451: "Unavailable For Legal Reasons",
            499: "Client Closed Request",

            # Server Error.
            500: "Internal Server Error",
            501: "Not Implemented",
            502: "Bad Gateway",
            503: "Service Unavailable",
            504: "Gateway Timeout",
            505: "Http Version Not Supported",
            506: "Variant Also Negotiates",
            507: "Insufficient Storage",
            509: "Bandwidth Limit Exceeded",
            510: "Not Extended",
            511: "Network Authentication Required",
        }
        return codes[self.value]

    @property
    def code(self):
        return self.value




# Decorators Request Methods
from bravaweb.utils import RequestMethod


def TestMethod(required, cls, f, **args):
    if required == cls.enviroment.method:
        return f(cls, **args)
    else:
        return cls.NotAllowed()


def post(f):
    def method_decorator(cls, **args) -> f:
        return TestMethod(RequestMethod.Post, cls, f, **args)
    return method_decorator


def get(f):
    def method_decorator(cls, **args) -> f:
        return TestMethod(RequestMethod.Get, cls, f, **args)
    return method_decorator


def put(f):
    def method_decorator(cls, **args) -> f:
        return TestMethod(RequestMethod.Put, cls, f, **args)
    return method_decorator


def delete(f):
    def method_decorator(cls, **args) -> f:
        return TestMethod(RequestMethod.Delete, cls, f, **args)
    return method_decorator




from mako.template import Template
from mako.lookup import TemplateLookup

# configuration
import configuration
import os

from bravaweb.response.lib.object import ResponseObject


class Html(ResponseObject):

    def FileName(self, **args):
        route = args["route"]
        action = args["action"]
        route[0] = "views"
        file = "{}/{}.html".format("/".join(route), action)
        if not os.path.exists("{}/{}".format(configuration.api.directory, file)):
            file = "views/shared/default.html"
        return file

    def Response(self, data, **args):

        _data = data

        _mylookup = TemplateLookup(directories=[''])

        _filename = self.FileName(**args)
        #args["template"] if "template" in args else "views/errors/default.html"

        template = Template(filename=_filename, lookup=_mylookup, input_encoding=configuration.api.encoding,
                            output_encoding=configuration.api.encoding, default_filters=['decode.{}'.format(configuration.api.encoding)], encoding_errors='replace')
        return template.render(data=_data)



# -*- coding: utf-8 -*-

import configuration

from enum import Enum


class ResponseType(Enum):

    # Application
    Edi = 1
    Edifact = 2
    JavaScript = 3
    OctetStream = 4
    Ogg = 5
    Pdf = 6
    Xhtml = 7
    Flash = 8
    Json = 9
    Ld_Json = 10
    Xml = 11
    Zip = 12

    # AUDIO
    MpegAudio = 13
    Wma = 14
    RealAudio = 15
    Wav = 16

    # IMAGE
    Gif = 17
    Jpg = 18
    Png = 19
    Tiff = 20
    MsIcon = 21
    Icon = 22
    Djvu = 23
    Svg = 24

    # MULTIPART
    MultipartMixex = 25
    MultipartAlternative = 26
    MultipartRelated = 27

    # PLAIN TEXT
    Css = 28
    Csv = 29
    Html = 30
    TextJavaScript = 31
    TextPlain = 32
    TextXml = 33

    # video
    MpegVideo = 34
    Mp4 = 35
    Quicktime = 36
    Wmv = 37
    MsVideo = 38
    Flv = 39
    Webm = 40

    # VND
    OpenDocument = 41
    OpenSpreadSheet = 42
    OpenPresentation = 43
    OpenGraphics = 44
    MsExcel = 45
    MsExcelSheet = 46
    MsPowerpoint = 47
    MsPowerpointPresentation = 48
    MsWord = 49
    MsWord_document = 50
    MozillaXul = 51

    def __str__(self):
        return "Content-Type: %s" % self.description

    @property
    def description(self):
        codes = {
            # APPLICATION
            1: "application/EDI-X12",
            2: "application/EDIFACT",
            3: "application/javascript; charset={charset}",
            4: "application/octet-stream",
            5: "application/ogg",
            6: "application/pdf",
            7: "application/xhtml+xml; charset={charset}",
            8: "application/x-shockwave-flash",
            9: "application/json; charset={charset}",
            10: "application/ld+json; charset={charset}",
            11: "application/xml; charset={charset}",
            12: "application/zip",
            # AUDIO
            13: "audio/mpeg",
            14: "audio/x-ms-wma",
            15: "audio/vnd.rn-realaudio",
            16: "audio/x-wav",
            # IMAGE
            17: "image/gif",
            18: "image/jpeg",
            19: "image/png",
            20: "image/tiff",
            21: "image/vnd.microsoft.icon",
            22: "image/x-icon",
            23: "image/vnd.djvu",
            24: "image/svg+xml",
            # MULTIPART
            25: "multipart/mixed",
            26: "multipart/alternative",
            27: "multipart/related",
            # PLAIN TEXT
            28: "text/css; charset={charset}",
            29: "text/csv; charset={charset}",
            30: "text/html; charset={charset}",
            31: "text/javascript; charset={charset}",
            32: "text/plain; charset={charset}",
            33: "text/xml; charset={charset}",
            # VIDEO
            34: "video/mpeg",
            35: "video/mp4",
            36: "video/quicktime",
            37: "video/x-ms-wmv",
            38: "video/x-msvideo",
            39: "video/x-flv",
            40: "video/webm",
            # VND
            41: "application/vnd.oasis.opendocument.text",
            42: "application/vnd.oasis.opendocument.spreadsheet",
            43: "application/vnd.oasis.opendocument.presentation",
            44: "application/vnd.oasis.opendocument.graphics",
            45: "application/vnd.ms-excel",
            46: "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
            47: "application/vnd.ms-powerpoint",
            48: "application/vnd.openxmlformats-officedocument.presentationml.presentation",
            49: "application/msword",
            50: "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
            51: "application/vnd.mozilla.xul+xml",
        }
        return codes[self.value].format(charset=configuration.api.encoding)

    @property
    def extension(self):
        exts = {
            # APPLICATION
            1: "",
            2: "",
            3: "js",
            4: "bin",
            5: "ogg",
            6: "pdf",
            7: "xhtml",
            8: "swf",
            9: "json",
            10: "json",
            11: "xml",
            12: "zip",
            # AUDIO
            13: "mpeg",
            14: "wma",
            15: "ra",
            16: "wav",
            # IMAGE
            17: "gif",
            18: "jpg",
            19: "png",
            20: "tiff",
            21: "ico",
            22: "ico",
            23: "djvu",
            24: "svg",
            # MULTIPART
            25: "",
            26: "",
            27: "",
            # PLAIN TEXT
            28: "css",
            29: "csv",
            30: "html",
            31: "js",
            32: "txt",
            33: "xml",
            # VIDEO
            34: "mpeg",
            35: "mp4",
            36: "mov",
            37: "wmv",
            38: "avi",
            39: "flv",
            40: "webm",
            # VND
            41: "odt",
            42: "ods",
            43: "pdp",
            44: "odg",
            45: "xlsx",
            46: "ods",
            47: "ppt",
            48: "ppt",
            49: "docx",
            50: "docx",
            51: "xml",
        }
        return exts[self.value]
